"""AMPQ support."""
import asyncio
import pickle
import typing as t
import uuid
from importlib import import_module

import aioamqp

from . import logger
from .utils import AsyncMixin


class Queue(AsyncMixin):

    """Support message queue."""

    defaults = dict(
        host="localhost",
        port=None,
        login="guest",
        password="guest",
        virtualhost="/",
        queue="donald",
        exchange_name="",
    )

    _started: bool = False
    _connected: bool = False
    _loop: t.Optional[asyncio.AbstractEventLoop] = None

    transport = None
    protocol: t.Optional[aioamqp.protocol.AmqpProtocol] = None
    channel: t.Optional[aioamqp.channel.Channel] = None

    def __init__(self, master: "Donald", **params):
        """Initialize the queue."""
        self.params = dict(self.defaults, **params)
        self.master = master

    def init_loop(self, loop: t.Optional[asyncio.AbstractEventLoop]):
        """Bind to given loop."""
        if not self._started:
            self._loop = loop or asyncio.get_event_loop()

    def is_connected(self) -> bool:
        """Check that the queue is connected."""
        return self._connected

    async def start(
        self, listen: bool = True, loop: asyncio.AbstractEventLoop = None
    ) -> bool:
        """Connect and start listen the message queue."""
        if self.master.params["fake_mode"]:
            return True

        logger.warning("Start Donald Queue")

        self.init_loop(loop)
        self._started = True

        await self.connect()
        if listen:
            channel = t.cast(aioamqp.channel.Channel, self.channel)
            await channel.basic_consume(self.listen, queue_name=self.params["queue"])

        return self._started

    async def stop(self, *args, **kwargs):
        """Stop listeners."""
        if not self.is_connected() or self.is_closed() or not self._started:
            return False

        self._started = False

        await self.protocol.close(no_wait=True)
        self.transport.close()
        self._connected = False

        logger.warning("Donald Queue is stopped")

    async def __aenter__(self) -> "Queue":
        """Support usage as a context manager."""
        await self.start()
        return self

    async def __aexit__(self, *args):
        """Support usage as a context manager."""
        await self.stop()

    async def connect(self) -> bool:
        """Connect to queue."""
        if not self._connected:
            logger.warning("Connect to queue: %r", self.params)
            try:
                self.transport, self.protocol = await asyncio.wait_for(
                    aioamqp.connect(
                        loop=self._loop, on_error=self.on_error, **self.params
                    ),
                    timeout=10,
                )
                self.channel = await self.protocol.channel()
                await self.channel.queue_declare(
                    queue_name=self.params["queue"], durable=True
                )
                await self.channel.basic_qos(
                    prefetch_count=1, prefetch_size=0, connection_global=False
                )
                self._connected = True

            except Exception as exc:
                self.on_error(exc)
                raise

        return self._connected

    def on_error(self, exc: Exception):
        """Reconnect on errors."""
        self._connected = False
        if self._started and not self.is_closed():
            logger.error(exc)
            self.loop.call_later(1, asyncio.create_task, self.connect())

    def submit(self, func: t.Callable, *args, **kwargs) -> asyncio.Future:
        """Submit to the queue."""
        if self.master.params["fake_mode"]:
            return self.master.submit(func, *args, **kwargs)

        logger.info("Submit to queue: '%s'", func.__qualname__)

        payload = pickle.dumps((func, args, kwargs))
        properties = dict(delivery_mode=2, message_id=str(uuid.uuid4()))

        if not self._connected:
            logger.warning("Donald Queue is not connected")
            fut: asyncio.Future = asyncio.Future()
            fut.set_result(False)
            return fut

        channel = t.cast(aioamqp.channel.Channel, self.channel)
        return asyncio.create_task(
            channel.basic_publish(
                payload=payload,
                exchange_name=self.params["exchange_name"],
                routing_key=self.params["queue"],
                properties=properties,
            )
        )

    async def listen(
        self,
        channel: aioamqp.channel.Channel,
        body: bytes,
        envelope: aioamqp.envelope.Envelope,
        properties: aioamqp.properties.Properties,
    ):
        """Run tasks from self.queue."""
        func, args, kwargs = pickle.loads(body)
        if isinstance(func, str):
            mod_name, _, func = func.rpartition(".")
            try:
                mod = import_module(mod_name)
                func = getattr(mod, func)
            except (ImportError, AttributeError) as exc:
                logger.error(exc)
                return False

        self.master.submit_nowait(func, *args, **kwargs)

        if channel:
            await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)


if t.TYPE_CHECKING:
    from .core import Donald  # noqa



"""Donald tasks' manager for Asyncio."""

import logging

logger = logging.getLogger(__name__)
handle = logging.StreamHandler()
handle.setFormatter(
    logging.Formatter(
        "[%(asctime)-15s] [%(process)d] [%(levelname)s] %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )
)
logger.addHandler(handle)


from .core import CronTab as crontab
from .core import Donald, run_donald

__all__ = "crontab", "Donald", "run_donald"

__version__ = "0.13.0"
__project__ = "Donald"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "BSD"



"""Do the work."""

import asyncio
import signal
from contextlib import contextmanager
from functools import partial
from queue import Empty

from . import logger
from .utils import create_task


def run_worker(rx, tx, params):
    """Create and run a worker inside a process."""
    worker = Worker(rx, tx, params)
    return worker.run()


class Worker:

    """Put a job into a separate process."""

    def __init__(self, rx, tx, params):
        """Initialize the worker."""
        self.rx = rx
        self.tx = tx
        self.params = params
        self.tasks = 0
        self.running = False

        if params["sentry"]:
            from sentry_sdk import init

            init(**params["sentry"])

    def run(self):
        """Wait for a command and do the job."""
        logger.setLevel(self.params["loglevel"].upper())
        loop = asyncio.events.new_event_loop()
        asyncio.events.set_event_loop(loop)

        def stop():
            self.running = False

        loop.add_signal_handler(signal.SIGINT, stop)
        loop.add_signal_handler(signal.SIGTERM, stop)

        loop.run_until_complete(self.listen())

    async def listen(self):
        """Listen for tasks and run."""
        logger.info("Start worker: loop %s", id(asyncio.get_event_loop()))
        await self.handle("on_start")
        self.running = True

        rx = self.rx
        params = self.params

        # Mark self as a ready
        self.tx.put(True)

        while self.running:
            if self.tasks < params["max_tasks_per_worker"]:
                try:
                    message = rx.get(block=False)
                    if message is None:
                        self.running = False
                        break

                    ident, func, args, kwargs = message
                    logger.info("Run: '%s'", func.__qualname__)
                    task = create_task(func, args, kwargs)
                    task.add_done_callback(partial(self.done, ident))
                    self.tasks += 1

                except Empty:
                    pass

            await asyncio.sleep(1e-2)

        # Stop the runner
        logger.info("Stop worker")
        await self.handle("on_stop")

        tasks = [
            task for task in asyncio.all_tasks() if task is not asyncio.current_task()
        ]
        if tasks:
            await asyncio.sleep(1)

        for task in tasks:
            task.cancel()

    def done(self, ident, task):
        """Send the task's result back to main."""
        with catch_exc(self):
            try:
                res = task.exception()
                if res is not None:
                    raise res

                res = task.result()

            except (asyncio.CancelledError, asyncio.InvalidStateError) as res:
                raise res

        self.tasks -= 1
        if ident:
            self.tx.put((ident, res))

    async def handle(self, etype):
        """Run handlers."""
        for (func, args, kwargs) in self.params.get(etype, []):
            with catch_exc(self):
                await create_task(func, args, kwargs)


@contextmanager
def catch_exc(worker: Worker):
    """Catch exceptions."""
    try:
        yield worker
    except BaseException as exc:
        logger.exception(exc)
        if worker.params["sentry"]:
            from sentry_sdk import capture_exception

            capture_exception(exc)



"""Utils."""

import asyncio as aio
import os
from inspect import iscoroutinefunction


class AsyncMixin:

    """Working with an event loop."""

    def is_running(self):
        """Is the related loop working."""
        return self._loop and self._loop.is_running()

    def is_closed(self):
        """Is the related loop closed."""
        return not self._loop or self._loop.is_closed()

    @property
    def loop(self):
        """Get the binded loop."""
        return self._loop


class FileLock(object):

    """Simplest filelock implementation."""

    class Error(Exception):

        """The lock's error."""

        pass

    def __init__(self, fname, timeout=None, force=False):
        """Initialize the lock."""
        self.fname = fname
        self.fh = None
        self.flags = os.O_CREAT | os.O_RDWR
        for flag in ("O_EXLOCK", "O_NOINHERIT"):
            self.flags |= getattr(os, flag, 0)

    def acquire(self):
        """Acquire the lock."""
        if os.path.exists(self.fname):
            raise self.Error()
        self.fh = os.open(self.fname, self.flags)
        os.write(self.fh, bytes(os.getpid()))

    def release(self, silent=False):
        """Release the lock."""
        try:
            if self.fh:
                os.close(self.fh)
            os.remove(self.fname)
        except OSError:
            if not silent:
                raise

    def __enter__(self):
        """Enter in the context."""
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        """Exit from the context."""
        self.release()


def create_task(func, args, kwargs):
    """Create a task from the given function."""
    if iscoroutinefunction(func):
        corofunc = func

    else:

        async def corofunc(*args, **kwargs):
            return func(*args, **kwargs)

    return aio.create_task(corofunc(*args, **kwargs))



"""CLI Support."""

from importlib import import_module

import click

from .core import Donald, run_donald


@click.command()
@click.option("--listen", help="Listen the given AMPQ URL.")
@click.option(
    "--workers", default=Donald.defaults["num_workers"], help="Number of workers."
)
@click.option("--loglevel", default=Donald.defaults["loglevel"], help="Logging level.")
@click.option("--module", help="Load module.")
def main(workers=4, loglevel="info", module=None, **kw):
    """Run Donald server."""
    donald = Donald(num_workers=workers, loglevel=loglevel)

    if module:
        try:
            click.echo("Import module: %s" % module)
            import_module(module)
        except ImportError:
            click.echo("Invalid module: %s" % module)

    run_donald(donald)


if __name__ == "__main__":
    main()



"""Setup the package."""

# Parse requirements
# ------------------

import pathlib

import pkg_resources


def parse_requirements(path: str) -> "list[str]":
    with pathlib.Path(path).open() as requirements:
        return [str(req) for req in pkg_resources.parse_requirements(requirements)]


# Setup package
# -------------

from setuptools import setup

setup(
    install_requires=parse_requirements("requirements/requirements.txt"),
    extras_require={
        "tests": parse_requirements("requirements/requirements-tests.txt"),
        "build": ["bump2version", "wheel"],
    },
)



"""Donald Asyncio Tasks."""

import asyncio
import datetime
import multiprocessing as mp
import signal
import typing as t
from functools import wraps
from queue import Empty

from crontab import CronTab

from . import logger
from .queue import Queue
from .utils import AsyncMixin, FileLock, create_task
from .worker import run_worker


class Donald(AsyncMixin):

    """I'am on Donald."""

    defaults: t.Dict[str, t.Any] = {
        # Run tasks imediatelly in the same process/thread
        "fake_mode": False,
        # Number of workers
        "num_workers": mp.cpu_count() - 1,
        # Maximum concurent tasks per worker
        "max_tasks_per_worker": 100,
        # Ensure that the Donald starts only once
        "filelock": None,
        # logging level
        "loglevel": "INFO",
        # Start handlers
        "on_start": [],
        # Stop handlers
        "on_stop": [],
        # AMQP params
        "queue_name": "donald",
        "queue_params": {},
        # Sentry options ({'dsn': '...'})
        "sentry": None,
    }

    crontab = CronTab

    _loop: t.Optional[asyncio.AbstractEventLoop] = None
    _started = False

    rx: t.Optional[mp.Queue] = None
    tx: t.Optional[mp.Queue] = None
    workers: t.Optional[t.Tuple[mp.context.SpawnProcess, ...]] = None
    listener: t.Optional[asyncio.Task] = None

    def __init__(
        self, on_start: t.Callable = None, on_stop: t.Callable = None, **params
    ):
        """Initialize donald parameters."""
        self.params = dict(self.defaults, **params)

        logger.setLevel(self.params["loglevel"].upper())
        logger.propagate = False

        self.lock = FileLock(self.params["filelock"])
        self.schedules: t.List = []
        self.waiting: t.Dict[int, asyncio.Future] = {}

        self.queue = Queue(
            self, **dict(self.params["queue_params"], queue=self.params["queue_name"])
        )

    def __str__(self) -> str:
        """Representate as a string."""
        return f"Donald [{self.params['num_workers']}]"

    async def start(self, loop: asyncio.AbstractEventLoop = None):
        """Start workers.

        :returns: A coroutine
        """
        logger.warning("Start Donald: loop %s", id(asyncio.get_event_loop()))
        self._loop = loop or asyncio.get_event_loop()
        self.queue.init_loop(loop)

        if self.params["fake_mode"]:
            return True

        if self.params["filelock"]:
            try:
                self.lock.acquire()
            except self.lock.Error:
                logger.warning("Donald is locked. Exit.")
                return

        ctx = mp.get_context("spawn")
        self.rx = ctx.Queue()
        self.tx = ctx.Queue()

        self.workers = tuple(
            ctx.Process(target=run_worker, args=(self.rx, self.tx, self.params))
            for _ in range(max(self.params["num_workers"], 1))
        )

        # Start workers
        for wrk in self.workers:
            wrk.start()
            self.tx.get(block=True)

        # Start listener
        self.listener = asyncio.create_task(self.listen())

        # Start schedulers
        for idx, schedule in enumerate(self.schedules):
            logger.info("Schedule '%s'", schedule.__qualname__)
            self.schedules[idx] = asyncio.create_task(schedule())

        # Mark self started
        self._started = True

        return True

    async def stop(self, *args, **kwargs):
        """Stop workers. Disconnect from queue. Cancel schedules.

        The method could be called syncroniously too.

        :returns: A future
        """
        if self.is_closed() or not self._started:
            return

        logger.warning("Stoping Donald")

        if self.params["filelock"]:
            self.lock.release(silent=True)

        # Stop runner if exists
        if self.listener:
            self.listener.cancel()

        # Stop schedules
        for task in self.schedules:
            task.cancel()

        # Stop workers
        for wrk in self.workers:
            self.rx.put(None)

        for wrk in self.workers:
            wrk.join(1)
            wrk.terminate()

        self.rx.close()
        self.tx.close()

        self._started = False

        if self.listener and not self.listener.done():
            await asyncio.sleep(1e-2)

        logger.warning("Donald is stopped")
        return True

    async def __aenter__(self) -> "Donald":
        """Support usage as a context manager."""
        await self.start()
        return self

    async def __aexit__(self, *args):
        """Support usage as a context manager."""
        await self.stop()

    def __submit(
        self, fut: t.Optional[asyncio.Future], func: t.Callable, *args, **kwargs
    ) -> t.Optional[asyncio.Future]:
        """Submit the given task to workers."""
        if not callable(func):
            raise ValueError("Invalid task: %r" % func)

        if not self._started:
            raise RuntimeError("Donald is not started yet")

        logger.debug("Submit: '%s'", func.__qualname__)
        ident = None
        if fut:
            ident = id(fut)
            self.waiting[ident] = fut

        # Send task to workers
        rx = t.cast(mp.Queue, self.rx)
        rx.put((ident, func, args, kwargs))
        return fut

    async def listen(self):
        """Wait for a result and process."""
        tx = self.tx
        waiting = self.waiting
        while True:
            if not waiting:
                await asyncio.sleep(1e-2)
                continue

            try:
                ident, res, *args = tx.get(block=False)
                fut = waiting.pop(ident, None)
                if fut:
                    if isinstance(res, Exception):
                        fut.set_exception(res)
                    else:
                        fut.set_result(res)

                # Leave the objects from the closure
                fut = ident = res = None

            except Empty:
                pass

            await asyncio.sleep(0)

    def submit(self, func: t.Callable, *args, **kwargs) -> asyncio.Future:
        """Submit a task to workers."""
        if self.params["fake_mode"]:
            return create_task(func, args, kwargs)

        fut = self.loop.create_future()
        res = self.__submit(fut, func, *args, **kwargs)
        res = t.cast(asyncio.Future, res)
        return res

    def submit_nowait(self, func: t.Callable, *args, **kwargs):
        """Submit a task to workers."""
        if self.params["fake_mode"]:
            create_task(func, args, kwargs)

        else:
            self.__submit(None, func, *args, **kwargs)

    def schedule(
        self,
        interval: t.Union[int, float, datetime.timedelta, CronTab],
        *args,
        **kwargs,
    ):
        """Add func to schedules. Use this as a decorator.

        Run given func/coro periodically.
        """
        if callable(interval):
            raise RuntimeError("@donald.schedule(interval) should be used.")

        if isinstance(interval, datetime.timedelta):
            timer = interval.total_seconds

        elif isinstance(interval, CronTab):
            timer = lambda: interval.next(default_utc=True)  # type: ignore

        else:
            timer = lambda: float(interval)  # type: ignore

        def wrapper(func):
            async def catcher():
                try:
                    await self.submit_nowait(func, *args, **kwargs)
                except Exception:
                    pass

            @wraps(func)
            async def scheduler():
                while True:
                    sleep = max(timer(), 0.01)
                    logger.info("Next '%s' in %0.2f s", func.__qualname__, sleep)
                    await asyncio.sleep(sleep)
                    asyncio.create_task(catcher())

            self.schedules.append(scheduler)
            return func

        return wrapper

    async def run(self, tick: int = 60):
        """Keep asyncio busy."""
        while self._started:
            logger.info("Donald is running")
            await asyncio.sleep(tick)

    def on_start(self, func: t.Callable, *args, **kwargs) -> t.Callable:
        """Register start handler."""
        self.params["on_start"].append((func, args, kwargs))
        return func

    def on_stop(self, func: t.Callable, *args, **kwargs) -> t.Callable:
        """Register stop handler."""
        self.params["on_stop"].append((func, args, kwargs))
        return func


def run_donald(donald, queue=True):
    """Help to run donald."""
    loop = asyncio.get_event_loop()

    async def stop_donald():
        if queue:
            await donald.queue.stop()
        await donald.stop()
        loop.stop()

    def handle_signal(loop):
        """Stop donald before exit."""
        loop.remove_signal_handler(signal.SIGTERM)
        loop.remove_signal_handler(signal.SIGINT)
        loop.create_task(stop_donald())

    loop.add_signal_handler(signal.SIGTERM, handle_signal, loop)
    loop.add_signal_handler(signal.SIGINT, handle_signal, loop)
    loop.run_until_complete(donald.start())
    if queue:
        loop.run_until_complete(donald.queue.start())

    try:
        loop.run_forever()
    finally:
        loop.close()
